# -*- coding: utf-8 -*-
__author__ = "Philipp D. Rohde"

from S2Spy.ShapeParser import ShapeParser
from S2Spy.sparql.SPARQLEndpoint import SPARQLEndpoint
from S2Spy.utils import fileManagement
from S2Spy.RuleBasedValidation import RuleBasedValidation
from S2Spy.utils.globals import PARSING_ORDER


class ShapeNetwork:

    def __init__(self, schemaDir, schemaFormat, endpointURL, useSelectiveQueries, maxSplitSize, outputDir, ORDERBYinQueries):
        self.shapes = ShapeParser().parseShapesFromDir(schemaDir, schemaFormat, useSelectiveQueries, maxSplitSize, ORDERBYinQueries)
        self.shapesDict = {shape.getId(): shape for shape in self.shapes}  # TODO: use only the dict?
        self.endpoint = SPARQLEndpoint(endpointURL)
        self.outputDirName = outputDir

    @staticmethod
    def get_node_order():
        """The execution order of SHACL2SPARQL is based on the order of files as returned by the file system."""
        return PARSING_ORDER

    def validate(self):
        """Execute the Validation of the Shape Network."""
        node_order = self.get_node_order()

        for s in self.shapes:
            s.computeConstraintQueries()

        instancesReport = self.getInstances(node_order)
        return instancesReport

    def getInstances(self, node_order):
        """
        Reports valid and violated constraints of the graph
        :param node_order:
        """
        RuleBasedValidation(
            self.endpoint,
            node_order,
            self.shapesDict,
            fileManagement.openFile(self.outputDirName, "validation.log"),
            fileManagement.openFile(self.outputDirName, "targets_valid.log"),
            fileManagement.openFile(self.outputDirName, "targets_violated.log"),
            fileManagement.openFile(self.outputDirName, "stats.txt"),
            fileManagement.openFile(self.outputDirName, "traces.csv")
        ).exec()
        return 'Go to log files in {} folder to see report'.format(self.outputDirName)



# -*- coding: utf-8 -*-


class DependencyGraphImpl:
    """ Maps a shape to an array of 2 sets:
            first set: positive shape references
            second set: negative shape references """

    def __init__(self, shapeMap):
        # e.g., {ActorShape = [[], []], MovieShape = [[ActorShape], []]}
        self.references = self.computeDependencyGraph(shapeMap)

    def computeDependencyGraph(self, shapeMap):
        references = {}
        for shape in shapeMap:
            references[shape] = self.getShapeRefs(shapeMap[shape], shapeMap)
        print("references", references)
        return references

    def getShapeRefs(self, shape, shapeMap):
        return [self.getPosShapeRefs(shape, shapeMap), self.getNegShapeRefs(shape, shapeMap)]

    def getPosShapeRefs(self, shape, shapeMap):
        return [shapeMap.get(r) for r in shape.getPosShapeRefs()[0] if shapeMap.get(r) is not None]

    def getNegShapeRefs(self, shape, shapeMap):
        return [shapeMap.get(r) for r in shape.getNegShapeRefs()[0] if shapeMap.get(r) is not None]


# -*- coding: utf-8 -*-
from S2Spy.utils import fileManagement
from S2Spy.utils.RuleBasedValidStats import RuleBasedValidStats
from S2Spy.core.RuleMap import RuleMap
from S2Spy.core.Literal import Literal
import time


class RuleBasedValidation:
    def __init__(self, endpoint, node_order, shapesDict, logOutput,
                 validTargetsOutput, invalidTargetsOutput, statsOutput, tracesOutput):
        self.endpoint = endpoint
        self.node_order = node_order
        self.shapesDict = shapesDict

        self.targetShapes = self.extractTargetShapes()
        self.targetShapePredicates = [s.getId() for s in self.targetShapes]

        self.logOutput = logOutput
        self.validOutput = validTargetsOutput
        self.violatedOutput = invalidTargetsOutput
        self.statsOutput = statsOutput
        self.stats = RuleBasedValidStats()

        self.tracesOutput = tracesOutput
        self.startOfVerification = time.time() * 1000.0
        self.tracesOutput.write("Target, #TripleInThatClass, TimeSinceStartOfVerification(ms)\n")

        self.evaluatedShapes = []  # used for the filtering queries
        self.prevEvalShapeName = None

    def exec(self):
        firstShapeEval = self.shapesDict[self.node_order.pop(0)]
        depth = 0  # evaluation order
        self.logOutput.write("Retrieving (initial) targets ...")
        targets = self.extractTargetAtoms(firstShapeEval)
        self.logOutput.write("\nTargets retrieved.")
        self.logOutput.write("\nNumber of targets:\n" + str(len(targets)))
        self.stats.recordInitialTargets(str(len(targets)))
        start = time.time()*1000.0
        state = EvalState(targets)
        self.validate(
            depth,
            state,
            firstShapeEval
        )
        finish = time.time()*1000.0
        elapsed = round(finish - start)
        self.stats.recordTotalTime(elapsed)
        print("Total execution time: ", str(elapsed), " ms")
        self.logOutput.write("\nMaximal (initial) number or rules in memory: " + str(self.stats.maxRuleNumber))
        self.writeTargetsToFile(state)
        self.stats.writeAll(self.statsOutput)

        fileManagement.closeFile(self.validOutput)
        fileManagement.closeFile(self.violatedOutput)
        fileManagement.closeFile(self.statsOutput)
        fileManagement.closeFile(self.logOutput)
        fileManagement.closeFile(self.tracesOutput)

    def getInstancesList(self, prevEvalShapeName):
        '''
        Returns the valid and invalid instances of the previous evaluated shape (if there is one).
        :param prevEvalShapeName: string containing the name of the shape of interest
        :return:
        '''
        if prevEvalShapeName is None:
            return [], []

        return self.shapesDict[prevEvalShapeName].bindings, self.shapesDict[prevEvalShapeName].invalidBindings

    def filteringTargetQuery(self, shape, bType, valList, invList):
        '''
        Local variables:
            maxSplitNumber: heuristic of maximum possible number of instances considered for using filtering queries
                            instead of the initial target query (currently hard-coded to 256)
            maxInstancesPerQuery: number from which the list is going to start being split because of the max
                            number of characters allowed in a query
        '''
        self.logOutput.write("\n" + bType + " instances shape: " + shape.id + " - child's (" + self.prevEvalShapeName + ")")
        return [shape.getTargetQuery()]

    def validTargetAtoms(self, shape, bType, prevValList, prevInvList):
        '''
        :param shape: current shape being evaluated
        :param bType: string containing instances type (either "valid" or "invalid")
        :param prevValList:
        :param prevInvList:
        :return:
        '''
        if len(prevValList) == 0:
            # no query to be evaluated since no new possible target literals can be found
            if len(prevInvList) > 0:
                shape.hasValidInstances = False
            return []  # no target literals retrieved
        else:
            query = self.filteringTargetQuery(shape, bType, prevValList, prevInvList)
            # when the lists are not split, the variable 'query' returns an array with one single query

            targetLiterals = set()
            for q in query:
                bindings = self.evalTargetQuery(shape, q)
                targetLiterals.update([Literal(shape.getId(), b["x"]["value"], True) for b in bindings])
            return targetLiterals

    def invalidTargetAtoms(self, shape, bType, prevValList, prevInvList):
        '''
        :param shape: current shape being evaluated
        :param bType: string containing instances type (either "valid" or "invalid")
        :param prevValList:
        :param prevInvList:
        :return:
        '''
        if len(prevInvList) == 0:
            return []
        else:
            query = self.filteringTargetQuery(shape, bType, prevValList, prevInvList)

            invalidBindings = set()
            qn = 0
            for q in query:
                bindings = self.evalTargetQuery(shape, q)
                print("^^^^^ bindings:", len(bindings))
                if qn == 0:
                    # update empty set
                    invalidBindings.update([Literal(shape.getId(), b["x"]["value"], True) for b in bindings])
                    qn += 1
                else:
                    invalidBindings.intersection([Literal(shape.getId(), b["x"]["value"], True) for b in bindings])
            print("invalidating...", len(invalidBindings))
            return invalidBindings

    # Extracts target atom from first evaluated shape
    def extractTargetAtoms(self, shape):
        return self.targetAtomsNaive(shape, shape.getTargetQuery())

    def evalTargetQuery(self, shape, query):
        if query is None:
            return []  # when a network has only some targets, a shape without a target class returns no new bindings
        self.logOutput.write("\nEvaluating query:\n" + query)
        startQ = time.time()*1000.0
        eval = self.endpoint.runQuery(
            shape.getId(),
            query,
            "JSON"
        )
        endQ = time.time()*1000.0
        self.logOutput.write("\nelapsed: " + str(endQ - startQ) + " ms\n")
        self.stats.recordQueryExecTime(endQ - startQ)
        self.stats.recordQuery()
        return eval["results"]["bindings"]

    def targetAtomsNaive(self, shape, targetQuery):
        '''
        :param shape:
        :param targetQuery: initial targetQuery is set in shape's definition file (json file)
        :return:
        '''
        bindings = self.evalTargetQuery(shape, targetQuery)
        return [Literal(shape.getId(), b["x"]["value"], True) for b in bindings]  # target literals

    def extractTargetShapes(self):
        return [s for name, s in self.shapesDict.items() if self.shapesDict[name].getTargetQuery() is not None]

    def checkPrevShapeValidationIsPending(self, prevEvalShapeName):
        if prevEvalShapeName is not None:
            prevShape = self.shapesDict[prevEvalShapeName]
            return len(prevShape.bindings) == 0 and len(prevShape.invalidBindings) == 0
        return False

    def validate(self, depth, state, focusShape):  # Algorithm 1 modified (SHACL2SPARQL)
        # termination condition 1: all targets are validated/violated  # does not apply anymore since we do not
        #if len(state.remainingTargets) == 0:                          # get all possible initial targets (considering
        #    return                                                    # all shapes of the network) from the beginning

        # termination condition 2: all shapes have been visited
        if len(state.visitedShapes) == len(self.shapesDict):
            for t in state.remainingTargets:
                self.registerTarget(t, True, depth, "not violated after termination", None, state)
            return

        self.logOutput.write("\n\n********************************")
        self.logOutput.write("\nStarting validation at depth: " + str(depth))

        self.evalShape(state, focusShape, depth)  # validate current selected shape

        # select next shape to be evaluated from the already defined list with the evaluation order
        if len(self.node_order) > 0:  # if there are more shapes to be evaluated
            self.nextEvalShape = self.shapesDict[self.node_order.pop(0)]
            self.evaluatedShapes.append(focusShape)
            if self.nextEvalShape.targetQuery is not None:
                self.logOutput.write("\n\n****************************************\n********************************")
                self.logOutput.write("\nRetrieving (next) targets ...")

                targets = self.extractTargetAtoms(self.nextEvalShape)
                state.remainingTargets.update(targets)
        else:
            self.nextEvalShape = None

        self.validate(depth + 1, state, self.nextEvalShape)

    def writeTargetsToFile(self, state):
        """
        Saves to local file the new validated target

        :param state:
        :return:
        """
        for valid in state.validTargets:
            self.validOutput.write(valid)

        for invalid in state.invalidTargets:
            self.violatedOutput.write(invalid)

        for trace in state.traces:
            self.tracesOutput.write(trace)

    def registerTarget(self, t, isValid, depth, logMessage, focusShape, state):
        fshape = ", focus shape " + focusShape.id if focusShape is not None else ""
        log = t.getStr() + ", depth " + str(depth) + fshape + ", " + logMessage + "\n"

        instance = "<" + t.getArg() + ">"

        if isValid:
            self.shapesDict[t.getPredicate()].bindings.add(instance)
            state.validTargets.add(log)
        else:
            self.shapesDict[t.getPredicate()].invalidBindings.add(instance)
            state.invalidTargets.add(log)

        targetType = "valid" if isValid else "violated"
        traceLog = targetType + ", " + \
                    str(state.tracesCount) + ", " + \
                    str(round(time.time() * 1000.0 - self.startOfVerification)) + "\n"

        state.traces.add(traceLog)
        state.tracesCount += 1


    def saturate(self, state, depth, s):
        negated = self.negateUnMatchableHeads(state, depth, s)
        inferred = self.applyRules(state, depth, s)

        if negated or inferred:
            self.saturate(state, depth, s)

    # INFER procedure performs 2 types of inferences:
    # 1. If the set of rules contains a rule and each of the rule bodies has already been inferred
    #    => head of the rule is inferred, rule dropped.
    # 2. If the negation of any rule body has already been inferred
    #    => this rule cannot be applied (rule head not inferred) so the entire entire rule is dropped.
    def applyRules(self, state, depth, s):
        retainedRules = RuleMap()                                                              # (2)
        freshLiterals = set(filter(lambda rule: rule is not None,                              # (4)
                                   [self._applyRules(head, bodies, state, retainedRules)
                                    for head, bodies in state.ruleMap.map.items()])            # (3)
                            )
        state.ruleMap = retainedRules
        state.assignment.update(freshLiterals)

        if len(freshLiterals) == 0:
            return False

        candidateValidTargets = {a for a in freshLiterals if a.getPredicate() in self.targetShapePredicates}
        remaining = set()
        for t in state.remainingTargets:
            if t in candidateValidTargets:
                if t.getIsPos():
                    #state.validTargets.add(t)
                    self.registerTarget(t, True, depth, "", s, state)
                else:
                    #state.invalidTargets.add(t)
                    self.registerTarget(t, False, depth, "", s, state)
            else:
                remaining.add(t)

        state.remainingTargets = remaining

        self.logOutput.write("\nRemaining targets: " + str(len(state.remainingTargets)))
        return True

    def _applyRules(self, head, bodies, state, retainedRules):
        for body in bodies:
            if body.issubset(state.assignment):  # means that there is at least one invalid rule
                return head

            matches = [a.getNegation() for a in body if a.getNegation() in state.assignment]  # ***
            if len(matches) == 0:  # no match
                retainedRules.addRule(head, body)  # not invalid
            return None

    # validateFocusShape
    def evalShape(self, state, s, depth):
        '''
        Saturate only if the current shape is connected in the network as a parent to a shape that was already
        evaluated and has any valid instances. This is,

        if the child node (previously evaluated shape) contains only invalid instances, all retrieved instances of the
        shape being currently evaluated were registered as invalid so there is no need to validate min/max constraints.

        :param state:
        :param s: focus shape
        :param depth:
        '''
        self.logOutput.write("\nEvaluating queries for shape " + s.getId())

        if s.hasValidInstances:
            self.evalConstraints(state, s)  # 'eval Disjunct'
            state.evaluatedPredicates.update(s.getPredicates())
            #self.saveRuleNumber(state)
            self.logOutput.write("\nStarting saturation ...")
            startS = time.time()*1000.0
            self.saturate(state, depth, s)
            endS = time.time()*1000.0
            self.stats.recordSaturationTime(endS - startS)
            self.logOutput.write("\nsaturation ...\nelapsed: " + str(endS - startS) + " ms")
        else:
            self.logOutput.write("\nNo saturation for shape ..." + s.getId())

        state.addVisitedShape(s)

        self.logOutput.write("\n\nValid targets: " + str(len(state.validTargets)))
        self.logOutput.write("\nInvalid targets: " + str(len(state.invalidTargets)))
        self.logOutput.write("\nRemaining targets: " + str(len(state.remainingTargets)))

    def saveRuleNumber(self, state):
        ruleNumber = state.ruleMap.getRuleNumber()
        self.logOutput.write("\nNumber of rules: " + str(ruleNumber))
        self.stats.recordNumberOfRules(ruleNumber)

    def filteringConstraintQuery(self, shape, templateQuery, prevValidInstances, prevInvalidInstances,
                         maxSplitNumber, maxInstancesPerQuery, qType):
        if self.prevEvalShapeName is not None \
                and len(prevValidInstances) > 0 and len(prevInvalidInstances) > 0 \
                and len(prevValidInstances) <= maxSplitNumber:
            VALUES_clauses = ""
            refPaths = "\n" if qType == "max" else ''
            formattedInstancesSplit = self.getFormattedInstances(prevValidInstances, "", maxInstancesPerQuery)
            for c in shape.constraints:
                if c.shapeRef == self.prevEvalShapeName:
                    var = " ?" + c.variables[0]
                    VALUES_clauses += "VALUES" + var + " {$instances$}\n"
                    if qType == "max":
                        focusVar = c.varGenerator.getFocusNodeVar()
                        refPaths += "?" + focusVar + " " + c.path + var + ".\n"

            return [templateQuery.replace("$to_be_replaced$",
                                           VALUES_clauses.replace("$instances$", sublist) + refPaths)
                    for sublist in formattedInstancesSplit]
        else:
            return [templateQuery.replace("$to_be_replaced$", "\n")]

    def evalConstraints(self, state, s):
        prevValList, prevInvList = self.getInstancesList(self.prevEvalShapeName)

        maxSplitNumber = s.maxSplitSize
        maxInstancesPerQuery = 80

        minQuery = self.filteringConstraintQuery(s, s.minQuery.getSparql(), prevValList, prevInvList,
                                             maxSplitNumber, maxInstancesPerQuery, "min")
        for queryStr in minQuery:
            self.evalQuery(state, s.minQuery, queryStr, s)

        for q in s.maxQueries:
            maxQuery = self.filteringConstraintQuery(s, q.getSparql(), prevValList, prevInvList,
                                                 maxSplitNumber, maxInstancesPerQuery, "max")
            for queryStr in maxQuery:
                self.evalQuery(state, q, queryStr, s)

    def evalQuery(self, state, q, query, s):
        self.logOutput.write("\n\nEvaluating query:\n" + query)
        startQ = time.time()*1000.0
        eval = self.endpoint.runQuery(q.getId(), query, "JSON")
        endQ = time.time()*1000.0
        self.logOutput.write("\nelapsed: " + str(endQ - startQ) + " ms\n")
        self.stats.recordQueryExecTime(endQ - startQ)

        bindings = eval["results"]["bindings"]  # list of obtained 'bindingsSet' from the endpoint

        self.logOutput.write("\nNumber of solution mappings: " + str(len(bindings)) + "\n")
        self.stats.recordNumberOfSolutionMappings(len(bindings))
        self.stats.recordQuery()
        queryRP = q.getRulePattern()
        qH = queryRP.getHead()
        qB = queryRP.getBody()
        qV = queryRP.getVariables()
        shapeRP = s.getRulePatterns()[0]
        sH = shapeRP.getHead()
        sB = shapeRP.getBody()
        sV = shapeRP.getVariables()
        bvars = set(bindings[0].keys()) if len(bindings) > 0 else set()

        startG = time.time() * 1000.0

        #for b in bindings:
            #self.evalBindingSet(state, b, queryRP.getVariables(), queryRP.getHead(), queryRP.getBody(), queryRP,
            #                    shapeRP.getVariables(), shapeRP.getHead(), shapeRP.getBody(), shapeRP)

        qRules = [state.ruleMap.addRule(
            Literal(
                qH.getPredicate(),
                binding[qH.getArg()]["value"],
                qH.getIsPos()),
            frozenset({Literal(a.getPredicate(),
                               binding[a.getArg()]["value"],
                               a.getIsPos())
                       for a in qB})
        ) for binding in bindings if bvars.issuperset(qV)]

        sRules = [state.ruleMap.addRule(
            Literal(
                sH.getPredicate(),
                binding[sH.getArg()]["value"],
                sH.getIsPos()),
            frozenset({Literal(a.getPredicate(),
                               binding[a.getArg()]["value"],
                               a.getIsPos())
                       for a in sB})
        ) for binding in bindings if bvars.issuperset(sV)]

        endG = time.time() * 1000.0

        self.stats.recordGroundingTime(endG - startG)
        self.logOutput.write("\nGrounding rules ... \nelapsed: " + str(endG - startG) + " ms\n")

    '''def evalBindingSet(self, state, bs, qV, qH, qB, qRP, sV, sH, sB, sRP):
        bindingVars = set(bs.keys())
        self._evalBindingSet(state, bs, bindingVars, qV, qH, qB, qRP)
        #for p in shapeRPs:  # for each pattern in the set of rule patterns (but a shape contains only one pattern
        self._evalBindingSet(state, bs, bindingVars, sV, sH, sB, sRP)

    def _evalBindingSet(self, state, bs, bindingVars, vars, head, body, rp):
        if bindingVars.issuperset(vars):
            state.ruleMap.addRule(
                rp.instantiateAtom(head, bs),
                frozenset({rp.instantiateAtom(a, bs) for a in body})
            )'''

    def negateUnMatchableHeads(self, state, depth, s):
        '''
        This procedure derives negative information
        For any (possibly negated) atom 'a' that is either a target or appears in some rule, we
        may be able to infer that 'a' cannot hold:
          if 'a' is not in state.assignment
          if the query has already been evaluated,
          and if there is not rule with 'a' as its head.
        Thus, in such case, 'not a' is added to state.assignment.

        :param state:
        :param depth:
        :param s:
        :return:
        '''
        ruleHeads = state.ruleMap.keySet()

        initialAssignmentSize = len(state.assignment)

        # first negate unmatchable body atoms (add not satisfied body atoms)
        allBodyAtoms = state.ruleMap.getAllBodyAtoms()
        state.assignment.update({a.getNegation() if a.getIsPos() else a
                                 for a in allBodyAtoms
                                 if not ((a.getPredicate() not in state.evaluatedPredicates)
                                          or (a.getAtom() in ruleHeads)
                                          or (a in state.assignment))}
                                )

        # then negate unmatchable targets
        remaining = set()
        for a in state.remainingTargets:
            if self.isSatisfiable(a, state, ruleHeads):
                remaining.add(a)
            else:
                #state.invalidTargets.add(a)
                state.assignment.add(a.getNegation())
                self.registerTarget(a, False, depth, "", s, state)

        state.remainingTargets = remaining

        return initialAssignmentSize != len(state.assignment)  # False when no new assignments are found

    #def getNegatedAtom(self, a):
    #    return a.getNegation() if a.getIsPos() else a

    def isSatisfiable(self, a, state, ruleHeads):
        return (a.getPredicate() not in state.evaluatedPredicates) \
                or (a.getAtom() in ruleHeads) \
                or (a in state.assignment)

class EvalState:
    def __init__(self, targetLiterals):
        self.remainingTargets = targetLiterals
        self.ruleMap = RuleMap()
        self.assignment = set()
        self.visitedShapes = set()
        self.evaluatedPredicates = set()
        self.validTargets = set()
        self.invalidTargets = set()
        self.traces = set()
        self.tracesCount = 0

    def addVisitedShape(self, s):
        self.visitedShapes.add(s)



# -*- coding: utf-8 -*-
__author__ = "Philipp D. Rohde"

from SPARQLWrapper import SPARQLWrapper, XML, JSON


class SPARQLEndpoint:
    """Implementation of a SPARQL endpoint; having an URL and the ability to run SPARQL queries."""

    class __SPARQLEndpoint:
        """Private class to allow simulation of a Singleton."""
        def __init__(self, endpointURL):
            self.endpointURL = endpointURL
            self.endpoint = SPARQLWrapper(endpointURL)
            self.endpoint.setReturnFormat(XML)

        def runQuery(self, queryId, queryString, format=None):
            self.endpoint.setQuery(queryString)
            if format == 'JSON':
                self.endpoint.setReturnFormat(JSON)
            return self.endpoint.query().convert()

    instance = None

    def __new__(cls, endpointURL):
        if not SPARQLEndpoint.instance:
            SPARQLEndpoint.instance = SPARQLEndpoint.__SPARQLEndpoint(endpointURL)
        return SPARQLEndpoint.instance

    def __getattr__(self, item):
        return getattr(self.instance, item)

    def __setattr__(self, key, value):
        return setattr(self.instance, key, value)



# -*- coding: utf-8 -*-
from S2Spy.DependencyGraphImpl import DependencyGraphImpl


class SchemaImpl:
    def __init__(self, shapes):
        self.shapes = shapes
        mapped = {}
        for s in shapes:
            mapped[s.getId()] = s
        self.shapeMap = mapped
        self.dependencyGraph = DependencyGraphImpl(self.shapeMap)

    def getShapes(self):
        return self.shapes



# -*- coding: utf-8 -*-
__author__ = "Monica Figuera and Philipp D. Rohde"

import os
import json
from S2Spy.sparql.SPARQLPrefixHandler import getPrefixString
from S2Spy.VariableGenerator import VariableGenerator
from S2Spy.constraints.MinMaxConstraint import MinMaxConstraint
from S2Spy.constraints.MinOnlyConstraint import MinOnlyConstraint
from S2Spy.constraints.MaxOnlyConstraint import MaxOnlyConstraint
from S2Spy.Shape import Shape
from S2Spy.utils.globals import PARSING_ORDER
from urllib.parse import urlparse


class ShapeParser:

    def __init__(self):
        return

    def parseShapesFromDir(self, path, shapeFormat, useSelectiveQueries, maxSplitSize, ORDERBYinQueries):
        fileExtension = self.getFileExtension(shapeFormat)
        filesAbsPaths = []
        # r=root, d=directories, f = files
        for r, d, f in os.walk(path):
            for file in f:
                if fileExtension in file:
                    filesAbsPaths.append(os.path.join(r, file))

        if not filesAbsPaths:
            raise FileNotFoundError(path + " does not contain any shapes of the format " + shapeFormat)

        if shapeFormat == "JSON":
            return [self.parseJson(p, useSelectiveQueries, maxSplitSize, ORDERBYinQueries) for p in filesAbsPaths]
        else:  # TODO: implement parsing of TTL format
            print("Unexpected format: " + shapeFormat)

    def getFileExtension(self, shapeFormat):
        if shapeFormat == "SHACL":
            return ".ttl"
        else:
            return ".json"  # dot added for convenience

    def parseJson(self, path, useSelectiveQueries, maxSplitSize, ORDERBYinQueries):
        targetQuery = None

        file = open(path, "r")
        obj = json.load(file)
        targetDef = obj.get("targetDef")

        name = obj["name"]
        PARSING_ORDER.append(name)
        id = name + "_d1"  # str(i + 1) but there is only one set of conjunctions
        constraints = self.parseConstraints(name, obj["constraintDef"]["conjunctions"], targetDef, id)

        includeSPARQLPrefixes = self.abbreviatedSyntaxUsed(constraints)
        prefixes = None
        if "prefix" in obj.keys():
            prefixes = obj["prefix"]
        prefix_string = ''
        if prefixes:
            prefix_string = "\n".join(["".join("PREFIX " + key + ": " + value) for (key, value) in prefixes.items()]) + "\n"
        referencedShapes = self.shapeReferences(obj["constraintDef"]["conjunctions"][0])

        if targetDef is not None:
            query = targetDef["query"]
            if query is not None:
                targetQuery = ''.join([prefix_string if includeSPARQLPrefixes else '', query])
            if urlparse(targetDef["class"]).netloc != '':  # if the target class is a url, add '<>' to it
                targetDef = '<' + targetDef["class"] + '>'
            else:
                targetDef = targetDef["class"]

        return Shape(name, targetDef, targetQuery, constraints, id, referencedShapes,
                     useSelectiveQueries, maxSplitSize, ORDERBYinQueries, includeSPARQLPrefixes, prefix_string)

    def abbreviatedSyntaxUsed(self, constraints):
        """
        Run after parsingConstraints.
        Returns false if the constraints' predicates are using absolute paths instead of abbreviated ones
        :param constraints: all shape constraints
        :return:
        """
        for c in constraints:
            if c.path.startswith("<") and c.path.endswith(">"):
                return False
        return True

    def shapeReferences(self, constraints):
        return {c.get("shape"): c.get("path") for c in constraints if c.get("shape") is not None}

    def parseConstraints(self, shapeName, array, targetDef, constraintsId):
        varGenerator = VariableGenerator()
        return [self.parseConstraint(varGenerator, array[0][i], constraintsId + "_c" + str(i + 1), targetDef) for i in range(len(array[0]))]

    def parseConstraint(self, varGenerator, obj, id, targetDef):
        min = obj.get("min")
        max = obj.get("max")
        shapeRef = obj.get("shape")
        datatype = obj.get("datatype")
        value = obj.get("value")
        path = obj.get("path")
        negated = obj.get("negated")

        oMin = None if (min is None) else int(min)
        oMax = None if (max is None) else int(max)
        oShapeRef = None if (shapeRef is None) else str(shapeRef)
        oDatatype = None if (datatype is None) else str(datatype)
        oValue = None if (value is None) else str(value)
        oPath = None if (path is None) else str(path)
        oNeg = True if (negated is None) else not negated  # True means it is a positive constraint

        if urlparse(path).netloc != '':  # if the predicate is a url, add '<>' to it
            oPath = '<' + path + '>'

        if oPath is not None:
            if oMin is not None:
                if oMax is not None:
                    return MinMaxConstraint(varGenerator, id, oPath, oMin, oMax, oNeg, oDatatype, oValue, oShapeRef, targetDef)
                return MinOnlyConstraint(varGenerator, id, oPath, oMin, oNeg, oDatatype, oValue, oShapeRef, targetDef)
            if oMax is not None:
                return MaxOnlyConstraint(varGenerator, id, oPath, oMax, oNeg, oDatatype, oValue, oShapeRef, targetDef)



# -*- coding: utf-8 -*-
__author__ = "Monica Figuera and Philipp D. Rohde"

from S2Spy.core.Literal import Literal


class Constraint:

    def __init__(self, id=None, isPos=None, satisfied=None, datatype=None, value=None, shapeRef=None, targetDef=None):
        self.id = id
        self.isPos = isPos
        self.satisfied = satisfied

        self.datatype = datatype
        self.value = value
        self.shapeRef = shapeRef
        self.target = targetDef

        self.variables = []

    def getDatatype(self):
        return self.datatype

    def getValue(self):
        return self.value

    def getShapeRef(self):
        return self.shapeRef

    def getTarget(self):
        return self.target

    def getId(self):
        return self.id

    def getIsPos(self):
        return self.isPos

    def generateVariables(self, varGenerator, type, numberOfVariables):
        vars = []
        for elem in range(numberOfVariables):
            vars.append(varGenerator.generateVariable(type))

        return vars

    def getVariables(self):
        return self.variables

    def computeRulePatternBody(self):
        return [Literal(self.shapeRef, v, self.isPos) for v in self.variables] \
                    if self.shapeRef is not None else []



# -*- coding: utf-8 -*-
import os
from S2Spy.ShapeNetwork import ShapeNetwork


class Eval:
    def __init__(self, args):
        """
        :type args: ...
        """
        self.outputDir = args.outputDir
        self.shapeFormat = "JSON"

        self.createOutputDir()
        schemaDir = args.d
        useSelectiveQueries = args.selective
        maxSplitSize = args.m
        ORDERBYinQueries = True
        self.network = ShapeNetwork(schemaDir, self.shapeFormat, args.endpoint, useSelectiveQueries,
                                    maxSplitSize, self.outputDir, ORDERBYinQueries)

        report = self.network.validate()  # run the evaluation of the SHACL constraints over the specified endpoint
        print("Report:", report)

    def createOutputDir(self):
        path = os.getcwd()
        os.makedirs(path + '/' + self.outputDir, exist_ok=True)



# -*- coding: utf-8 -*-
from S2Spy.core.RulePattern import RulePattern
from S2Spy.core.Literal import Literal
from S2Spy.core.Query import Query
from S2Spy.VariableGenerator import VariableGenerator
from S2Spy.constraints.Constraint import Constraint


class QueryGenerator:
    def __init__(self, shape):
        self.shape = shape

    def generateQuery(self, id, constraints, target, isSelective, includeORDERBY, includePrefixes, graph=None, subquery=None):
        # Only one max constraint per query is allowed, then 'constraints' arg contain only 1 element for the max case
        rp = self.computeRulePattern(constraints, id)

        builder = QueryBuilder(id, graph, subquery, rp.getVariables(), isSelective, target, constraints, includeORDERBY, self.shape.prefix_string)
        for c in constraints:
            builder.buildClause(c)

        return builder.buildQuery(rp, includePrefixes)

    @staticmethod
    def computeRulePattern(constraints, id):
        body = []
        for c in constraints:
            body = body + c.computeRulePatternBody()

        return RulePattern(
            Literal(id, VariableGenerator.getFocusNodeVar(), True),
            body
        )

    @staticmethod
    def generateLocalSubquery(graphName, posConstraints):
        localPosConstraints = [c for c in posConstraints if c.getShapeRef() is None]

        if len(localPosConstraints) == 0:
            return None  # Optional empty

        builder = QueryBuilder(
            "tmp",
            graphName,
            None,
            VariableGenerator.getFocusNodeVar()
        )

        for c in localPosConstraints:
            builder.buildClause(c)

        return builder.getSparql(False, True)

# mutable
# private class
class QueryBuilder:
    def __init__(self, id, graph, subquery, projectedVariables, isSelective=None, targetPath=None, constraints=None, includeORDERBY=None, prefix_string=''):
        self.id = id
        self.graph = graph
        self.subQuery = subquery
        self.projectedVariables = projectedVariables
        self.filters = []
        self.triples = []

        self.considerSelectivity = isSelective and targetPath is not None
        self.isSelective = isSelective
        self.targetPath = targetPath
        self.constraints = constraints
        self.includeORDERBY = includeORDERBY if includeORDERBY is not None else False
        self.prefix_string = prefix_string

    def addTriple(self, path, object):
        self.triples.append(
            "?" + VariableGenerator.getFocusNodeVar() + " " +
                path + " " +
                object + "."
        )

    def addDatatypeFilter(self, variable, datatype, isPos):
        s = self.__getDatatypeFilter(variable, datatype)
        self.filters.append(
            s if isPos else "!(" + s + ")"
        )

    def addConstantFilter(self, variable, constant, isPos):
        s = variable + " = " + constant
        self.filters.append(
            s if isPos else "!(" + s + ")"
        )

    def __getDatatypeFilter(self, variable, datatype):
        return "datatype(?" + variable + ") = " + datatype

    def getSparql(self, includePrefixes, isSubQuery):  # assuming optional graph
        if isSubQuery:  # creating the subquery
            return self.getQuery(False)

        prefixes = self.prefix_string if includePrefixes else ''
        selectiveClosingBracket = "}}" if self.considerSelectivity else ''
        outerQueryClosing = ''.join(["}\n" if self.subQuery is not None else '',
                                     "}" if self.getTriplePatterns() != '' and self.subQuery is not None else '',
                                     "}" if self.getTriplePatterns() != '' else ''])

        return ''.join([prefixes,
                        self.getSelective(),
                        self.getQuery(includePrefixes),
                        self.subQuery if self.subQuery is not None else '',
                        outerQueryClosing,
                        selectiveClosingBracket,
                        " ORDER BY ?" + VariableGenerator.getFocusNodeVar() if self.includeORDERBY else ''])

    def getQuery(self, includePrefixes):
        tempString = ""
        if includePrefixes:
            if "_pos" in self.id or "_max_" in self.id:
                # add VALUES clause to external query
                tempString = "$to_be_replaced$"

        triplePatterns = self.getTriplePatterns()
        if triplePatterns != '':
            return ''.join([self.getProjectionString(),
                            " WHERE {\n",
                            tempString,
                            triplePatterns,
                            "\n", "{\n" if self.subQuery is not None else ''])
        else:
            return ''

    def getSelective(self):
        if self.considerSelectivity:
            return "SELECT DISTINCT" + \
                   ", ".join(["?" + v for v in self.projectedVariables]) + " WHERE {\n" + \
                   "?" + VariableGenerator.getFocusNodeVar() + " a " + self.targetPath + " {\n"
        return ""

    def getProjectionString(self):
        return "SELECT DISTINCT " + \
               ", ".join(["?" + v for v in self.projectedVariables])

    def getTriplePatterns(self):
        tripleString = "\n".join(self.triples)

        if len(self.filters) == 0:
            return tripleString

        return tripleString + self.generateFilterString()

    def generateFilterString(self):
        if len(self.filters) == 0:
            return ""

        return "\nFILTER(\n" + \
                (self.filters[0] if len(self.filters) == 1 else " AND\n".join(self.filters)
                ) + ")"

    def addCardinalityFilter(self, variables):
        for i in range(0, len(variables)):
            for j in range(i + 1, len(variables)):
                self.filters.append("?" + variables[i] + " != ?" + variables[j])

    def buildClause(self, c):
        variables = c.getVariables()

        if isinstance(c, Constraint):
            path = c.path

            if c.getValue() is not None:        # means there is a existing reference to another shape
                self.addTriple(path, c.getValue())
                return

            for v in variables:
                self.addTriple(path, "?" + v)

        if c.getValue() is not None:
            self.addConstantFilter(
                    variables.iterator().next(),
                    c.getValue().get(),
                    c.getIsPos()
            )

        if c.getDatatype() is not None:
            for v in variables:
                self.addDatatypeFilter(v, c.getDatatype(), c.isPos())

        if len(variables) > 1:
            self.addCardinalityFilter(variables)

    def buildQuery(self, rulePattern, includePrefixes):
        return Query(
                self.id,
                rulePattern,
                self.getSparql(includePrefixes, False)
        )



class RuleMap:
    def __init__(self):
        self.map = {}

    def addRule(self, head, body):
        if self.map.get(head) is None:
            s = set()
            s.add(body)
            self.map[head] = s
        else:
            self.map[head].add(body)  # does not return a boolean value, therefore summing 1 to ruleNumber each time is not going to be accurate

    def getAllBodyAtoms(self):
        return set(frozenset().union(*set().union(*self.map.values())))

    def keySet(self):
        return set(self.map.keys())



# -*- coding: utf-8 -*-
__author__ = "Monica Figuera and Philipp D. Rohde"

from S2Spy.VariableGenerator import VariableType
from S2Spy.constraints.Constraint import Constraint


class MaxOnlyConstraint(Constraint):

    def __init__(self, varGenerator, id, path, max, isPos, datatype=None, value=None, shapeRef=None, targetDef=None):
        super().__init__(id, isPos, None, datatype, value, shapeRef, targetDef)
        self.varGenerator = varGenerator
        self.path = path
        self.min = -1
        self.max = max
        self.variables = self.computeVariables()

    def computeVariables(self):
        atomicConstraint = Constraint()
        return atomicConstraint.generateVariables(self.varGenerator, VariableType.VALIDATION, self.max + 1)

    @property
    def getMax(self):
        return self.max

    @property
    def getPath(self):
        return self.path



# -*- coding: utf-8 -*-
__author__ = "Philipp D. Rohde and Monica Figuera"

import itertools
from S2Spy.VariableGenerator import VariableGenerator
from S2Spy.core.Literal import Literal
from S2Spy.core.RulePattern import RulePattern
from S2Spy.sparql.QueryGenerator import QueryGenerator


class Shape:

    def __init__(self, id, targetDef, targetQuery, constraints, constraintsId, referencedShapes,
                 useSelectiveQueries, maxSplitSize, ORDERBYinQueries, includeSPARQLPrefixes, prefix_string):
        self.id = id
        self.constraints = constraints
        self.constraintsId = constraintsId
        self.targetDef = targetDef
        self.targetQuery = targetQuery  # Might be None
        self.rulePatterns = ()
        self.satisfied = None
        self.inDegree = None
        self.outDegree = None

        self.minQuery = ""
        self.maxQueries = ""
        self.predicates = []

        self.referencedShapes = referencedShapes
        self.bindings = set()
        self.invalidBindings = set()
        self.hasValidInstances = True

        self.useSelectiveQueries = useSelectiveQueries
        self.maxSplitSize = maxSplitSize
        self.ORDERBYinQueries = ORDERBYinQueries
        self.includePrefixes = includeSPARQLPrefixes
        self.prefix_string = prefix_string

    def getId(self):
        return self.id

    def getPredicates(self):
        return self.predicates

    def setDegree(self, inDegree, outDegree):
        self.inDegree = inDegree
        self.outDegree = outDegree

    def computePredicateSet(self, minQueryId, maxQueriesIds):
        '''
        Returns the ids of the queries for a shape
        '''

        return [self.id] + [self.id + "_d1"] + [minQueryId] + [q for q in maxQueriesIds]

    def computeTargetDef(self):
        #targets = SourceDescription.instance.get_classes(self.predicates)
        #for c in self.constraints:
        #    c.target = targets
        # fix: set target for constraints only (needed for the queries)
        # but not for the shape since it will interfere with the heuristics
        return None

#    def getPosShapeRefs(self):
#        return [d.getPosShapeRefs() for d in self.disjuncts]

#    def getNegShapeRefs(self):
#        return [d.getNegShapeRefs() for d in self.disjuncts]

#    def askViolations(self):
#        if self.targetDef is not None:  # not checking violations on shapes without target definitions
#            triple = re.findall(r'{.*}', self.targetDef)[0]  # *** considering only one target def
#            triple = triple[1:len(triple)-1]  # removed curly braces
#            triple = triple.strip().split()
#            target = triple[2]
#
#            minConstraints = self.disjuncts[0].minConstraints
#            for c in minConstraints:
#                c.violated = ASKQuery(c.path, target).evaluate("min", c.min)
#
#            maxConstraints = self.disjuncts[0].maxConstraints
#            for c in maxConstraints:
#                c.violated = ASKQuery(c.path, target).evaluate("max", c.max)

    def getTargetQuery(self):
        return self.targetQuery

    def getConstraints(self):
        return self.constraints

    def getNumberConstraints(self):
        """get the number of constraints belonging to this shape"""
        return len(self.constraints)

    def getShapeRefs(self):
        return [c.getShapeRef() for c in self.constraints if c.getShapeRef() is not None]

    def getRulePatterns(self):
        return self.rulePatterns

    def computeConstraintQueries(self):
        minConstraints = [c for c in self.constraints if c.min != -1]
        maxConstraints = [c for c in self.constraints if c.max != -1]
        queryGenerator = QueryGenerator(self)

        subquery = queryGenerator.generateLocalSubquery(None, minConstraints)

        # Build a unique set of triples (+ filter) for all min constraints
        minId = self.constraintsId + "_pos"
        self.minQuery = queryGenerator.generateQuery(
                minId,
                [c for c in minConstraints if c.getShapeRef() is not None],
                self.targetDef,
                self.useSelectiveQueries,
                self.ORDERBYinQueries,
                self.includePrefixes,
                None,
                subquery
        )

        # Build one set of triples (+ filter) for each max constraint
        maxIds = [self.constraintsId + "_max_" + str(i) for i in range(1, len(maxConstraints) + 1)]
        i = itertools.count()
        self.maxQueries = [queryGenerator.generateQuery(
                                        maxIds[next(i)],
                                        [c],
                                        self.targetDef,
                                        self.useSelectiveQueries,
                                        self.ORDERBYinQueries,
                                        self.includePrefixes,
                                        None,
                                        subquery) for c in maxConstraints]

        self.predicates = self.computePredicateSet(minId, maxIds)

        self.rulePatterns = self.computeRulePatterns()

    def computeRulePatterns(self):
        focusNodeVar = VariableGenerator.getFocusNodeVar()
        head = Literal(self.id, focusNodeVar, True)

        return [RulePattern(head, self.getDisjunctRPBody())]

    def getDisjunctRPBody(self):
        focusNodeVar = VariableGenerator.getFocusNodeVar()
        maxQueries = [Literal(s, focusNodeVar, False) for s in [q.getId() for q in self.maxQueries]]

        return [Literal(self.minQuery.id,
                        focusNodeVar,
                        True
                        )] + \
                maxQueries



# -*- coding: utf-8 -*-
from S2Spy.core.Literal import Literal


class RulePattern:
    # If a value for each variable is produced (by a solution mapping), then the rule pattern can be instantiated.
    # Note that it may be the case that these variables do not appear in the the body of the rule (because there is no constraint to propagate on these values, they only need to exist)

    def __init__(self, head, body):
        self.head = head
        self.literals = body

        self.variables = set([head.getArg()] + [a.getArg() for a in body if a is not None])

    def getHead(self):
        return self.head

    def getBody(self):
        return self.literals

    def instantiateAtom(self, a, bs):
        # given a binding with many possible projected variables, returns the atom that matches the variable
        return Literal(
                a.getPredicate(),
                bs[a.getArg()]["value"],  # arg instance, e.g., http://dbpedia.org/resource/Titanic_(1953_film),
                a.getIsPos()
        )

    def instantiateBody(self, bs):
        return frozenset({self.instantiateAtom(a, bs) for a in self.literals})

    def getVariables(self):
        return self.variables



# -*- coding: utf-8 -*-

class Literal:
    def __init__(self, pred, arg, isPos):
        self.pred = pred
        self.arg = arg
        self.isPos = isPos

    def __repr__(self):
        sign = "" if self.isPos else "!"
        return sign + self.pred + "(" + self.arg + ")"

    def __eq__(self, other):
        return isinstance(other, Literal) and \
               ((self.pred, self.arg, self.isPos) == (other.pred, other.arg, other.isPos))

    def __ne__(self, other):
        return not self.__eq__(other)

    def __hash__(self):
        return hash(self.__repr__())

    def getPredicate(self):
        return self.pred

    def getAtom(self):
        return self if self.isPos else self.getNegation()

    def getArg(self):
        return self.arg

    def getIsPos(self):
        return self.isPos

    def equals(self, o):
        if self == o:
            return True
        if o is None or type(self) != type(o):
            return False
        literal = o
        return self.isPos == literal.isPos and \
               self.pred == literal.pred and \
               self.arg == literal.arg

    def getNegation(self):
        return Literal(self.pred, self.arg, not self.isPos)

    def getStr(self):
        sign = "" if self.isPos else "!"
        return sign + self.getPredicate() + "(" + self.getArg() + ")"



from setuptools import setup

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    name='SHACL2SPARQLpy',
    version='1.0.0',
    packages=['S2Spy', 'S2Spy.constraints', 'S2Spy.core', 'S2Spy.sparql', 'S2Spy.utils'],
    license='MIT',
    author='Mónica Figuera, Philipp D. Rohde',
    author_email='philipp.rohde@tib.eu',
    url='https://github.com/SDM-TIB/S2Spy',
    download_url='https://github.com/SDM-TIB/S2Spy/archive/refs/tags/v1.0.0.tar.gz',
    description='Python reference implementation of SHACL2SPARQL',
    long_description=long_description,
    long_description_content_type="text/markdown",
    install_requires=['SPARQLWrapper>=1.8.5'],
    python_requires='>=3.6',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Intended Audience :: Science/Research'
      ]
)



# -*- coding: utf-8 -*-
import itertools
from enum import Enum

i = itertools.count()


class VariableGenerator:
    def __init__(self):
        self.index = 0

    @staticmethod
    def generateVariable(type):
        type = "p_"  # *** hardcoded
        return str(type) + str(next(i))

    @staticmethod
    def getFocusNodeVar():
        return "x"  # ***


class VariableType(Enum):
    VALIDATION = "p_"
    VIOLATION = "n_"  # not used

    @classmethod
    def prefix(cls):
        return cls.VALIDATION



# -*- coding: utf-8 -*-

class RuleBasedValidStats:
    def __init__(self):
        self.initialTargets = 0

        self.maxRuleNumber = 0
        self.totalSolutionMappings = 0
        self.maxSolutionMappings = 0

        self.totalQueryExectime = 0
        self.maxQueryExectime = 0
        self.totalGroundingTime = 0
        self.maxGroundingTime = 0
        self.totalSaturationTime = 0
        self.maxSaturationTime = 0
        self.numberOfQueries = 0

        self.totalTime = 0

    def writeAll(self, statsOutput):
        statsOutput.write("(initial) targets:\n" + str(self.initialTargets))
        statsOutput.write("\nmax number of solution mappings for a query:\n" + str(self.maxSolutionMappings))
        statsOutput.write("\ntotal number of solution mappings:\n" + str(self.totalSolutionMappings))
        statsOutput.write("\nmax number of rules in memory:\n" + str(self.maxRuleNumber))
        statsOutput.write("\nnumber of queries:\n" + str(self.numberOfQueries))
        statsOutput.write("\nmax exec time for a query:\n" + str(self.maxQueryExectime))
        statsOutput.write("\ntotal query exec time:\n" + str(self.totalQueryExectime))
        statsOutput.write("\nmax grounding time for a query:\n" + str(self.maxGroundingTime))
        statsOutput.write("\ntotal grounding time:\n" + str(self.totalGroundingTime))
        statsOutput.write("\nmax saturation time:\n" + str(self.maxSaturationTime))
        statsOutput.write("\ntotal saturation time:\n" + str(self.totalSaturationTime))
        statsOutput.write("\ntotal time:\n" + str(self.totalTime) + "\n")

    def recordInitialTargets(self, k):
        self.initialTargets = k

    def recordGroundingTime(self, ms):
        if ms > self.maxGroundingTime:
            self.maxGroundingTime = ms

        self.totalGroundingTime += ms

    def recordQueryExecTime(self, ms):
        if ms > self.maxQueryExectime:
            self.maxQueryExectime = ms

        self.totalQueryExectime += ms

    def recordSaturationTime(self, ms):
        if ms > self.maxSaturationTime:
            self.maxSaturationTime = ms

        self.totalSaturationTime += ms

    def recordNumberOfRules(self, k):
        if k > self.maxRuleNumber:
            self.maxRuleNumber = k

    def recordNumberOfSolutionMappings(self, k):
        if k > self.maxSolutionMappings:
            self.maxSolutionMappings = k

        self.totalSolutionMappings += k

    def recordTotalTime(self, ms):
        self.totalTime = ms

    def recordQuery(self):
        self.numberOfQueries += 1



# -*- coding: utf-8 -*-

class Query:
    def __init__(self, id, rulePattern, sparql):
        self.id = id
        self.rulePattern = rulePattern
        self.sparql = sparql

    def getId(self):
        return self.id

    def getRulePattern(self):
        return self.rulePattern

    def getSparql(self):
        return self.sparql
